#include <stdio.h>
#include "mmc.h"
#include "fat16.h"
#include "uart.h"

//________________________________________________________________________________________________________________________________________
// Module name:			fat16.c 
// Compiler used:		avr-gcc 3.4.5
// Last Modifikation:	23.05.2006
// Version:				1.8
// Authors:				Stephan Busker 	
// Description:			Source files for FAT16 implementation with read and write-access using AVR-Mikrocontrollers
//						Copyright (C) 2006 Stephan Busker
//........................................................................................................................................
// Functions:			unsigned char 	InitFat16(void)	
//						File 			*FileOpen(unsigned char *fname, char *mode)
//						char 			FileGetchar(FILE *file)
//						char 			FilePutChar(FILE *file,char c)
//						unsigned char   FileRead(File *file,unsigned char*,unsigned char)
//
// ext. functions:		extern unsigned char mmc_read_sector (unsigned long,unsigned char *);		
//						extern unsigned char mmc_write_sector (unsigned long,unsigned char *);
//........................................................................................................................................
//
// URL:					www.Mikro-Control.de
//________________________________________________________________________________________________________________________________________




//________________________________________________________________________________________________________________________________________
// 
// Global variables needed read- or write-acces to the FAT16- filesystem.
//				
//________________________________________________________________________________________________________________________________________

unsigned char 		SectorsPerCluster 		= 0;			// how many sectors does a cluster contain?
unsigned char		FatCopies 				= 0;			// Numbers of copies of the FAT
unsigned int		PossibleRootEntries 	= 0;			// Possible number of entries in the root directory.
unsigned int		SectorsPerFat 			= 0;			// how many sectors does a fat16 contain?
unsigned long		ReservedSectors 		= 0;			// Sectors reserved by the filesystem.
unsigned long		FirstPartitionSector 	= 0;			// Distance in sectors between the first partition and the master bootrecord.
unsigned long		FileAllocationTable 	= 0;			// pointer to the first FAT
unsigned long		RootDirectory 			= 0;			// Pointer to the rootdirectory of the first partition.
unsigned long		FirstDataCluster 		= 0;			// Pointer to the first cluster containing data (cluster0).


//File fpoint;												// memmory space allocation for the file that will be opend in main.
//unsigned char   FileBuffer[512];							// Buffer for read and write operation from or to the mmc.
struct DirEntry *DirectoryEntry;							// Pointer to an entry of the directory.
struct FatEntry *Fat;										// Pointer to an entry of the fat (next clusterposition).

unsigned char	*FileBuffer;								// Pointer to the sectorbuffer for reading or writing.
#if USE_MMC	


//________________________________________________________________________________________________________________________________________
// Funtion: 	InitFat16(void);
// 
// Description:	This function reads the Masterbootrecord and finds the position of the Volumebootrecord, the FAT and the Rootdirectory
//				and stores the information in global variables.
//________________________________________________________________________________________________________________________________________

unsigned char InitFat16(void)
{	
#if USE_MMC	

	mmc_read_sector((unsigned long)MBR_SECTOR,&FileBuffer[0]);				// Read the master bootrecord from mmc.	
	FirstPartitionSector= *((unsigned long*)&FileBuffer[0x1C6]);
	
	mmc_read_sector((unsigned long)FirstPartitionSector,&FileBuffer[0]);		// Read the volume bootrecord from mmc.	
	SectorsPerCluster	= FileBuffer[0x0D];
	FatCopies 			= FileBuffer[0x10];
	PossibleRootEntries = *((unsigned int*) &FileBuffer[0x11]);
	SectorsPerFat 		= *((unsigned int*) &FileBuffer[0x16]);
	ReservedSectors		= *((unsigned int*)&FileBuffer[0x0E]);
															// calculate the sectorpositon of the FAT, the Rootdirectory and the first Datacluster.
	FileAllocationTable	= (unsigned long)((unsigned long)FirstPartitionSector + (unsigned long)ReservedSectors);
	RootDirectory		= (unsigned long)((unsigned long)FileAllocationTable + (unsigned long)((unsigned long)SectorsPerFat*(unsigned long)FatCopies));	
	FirstDataCluster	= (unsigned long)((unsigned long)RootDirectory + ((unsigned long)(PossibleRootEntries>>4)));

  //  printf("Init:\n\rRoot: %lu FAT: %lu  1:%lu r:%lu\n\r",RootDirectory, FileAllocationTable, FirstPartitionSector, ReservedSectors);
#endif
	return(0);
}



//________________________________________________________________________________________________________________________________________
// Funtion: 	GetNextCluster(File *file);
// 
// Description:	This function finds the next datacluster of the file specified with File *file. 
//				
//________________________________________________________________________________________________________________________________________

unsigned int GetNextCluster(File *file)
{
	unsigned long fat_pointer			=	0;
	unsigned long fat_sector_offset	=	0;
	unsigned long ul_tmp				=  	0;
	
	fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));		// Calculate index of actual cluster in FAT.
	fat_sector_offset  /= SectorsPerCluster;														
	fat_sector_offset  += 2;													// In Fat16 clusterpositions have an offset of two.	
	fat_pointer = (fat_sector_offset%0x100);									// Calculate the clusterposition in the fat																
	fat_sector_offset   = (fat_sector_offset>>8);								// and the sectoroffset in relation to the beginning of the fat.

	mmc_read_sector((unsigned long)(FileAllocationTable + fat_sector_offset),&FileBuffer[0]);	
	file->sector_in_buffer = (FileAllocationTable + fat_sector_offset);			// Mark that new sector has been read.
	
	ul_tmp  = (unsigned long)FileBuffer[((fat_pointer << 1)+1)];				// Read next sector information from calculated clusterposition.
	ul_tmp  = (ul_tmp << 8);
	ul_tmp |= (unsigned long)FileBuffer[(fat_pointer << 1)];    
	ul_tmp -=2;																	// next datacluster is clusterposition in fat - 2.
	ul_tmp *= SectorsPerCluster;												// calculate sectorposition of new cluster
	ul_tmp += FirstDataCluster;													// in relation to first datacluster of the disk.
	
	file->cluster_pointer = (unsigned long) ul_tmp;							// continue reading the file at the beginning of new datacluster.

	return(0);
}




//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int FileReadSector(File *file);
// 
// Description:	This function reads cnt bytes into the FileBuffer. The number of databytes available in the buffer is returned. 
//				
//________________________________________________________________________________________________________________________________________


unsigned int FileReadSector(File *file)						 
{
	unsigned int  cnt		= 0;
	unsigned long temp1	= 0; 										

	if(file->filesize > 0)														// wen the end of the file is not reached, get the next character.
	{
		file->byte_index=0;														// reading at the beginning of new sector.
		if(file->sector_index >= SectorsPerCluster)							// When end of cluster is reached, the next datacluster has to be searched in the FAT.
		{
			file->sector_index = 0;												// start reading new cluster at first sector of the cluster.
			GetNextCluster(file);												// Sets the clusterpointer of the file to the next datacluster.
		}

		temp1  = (unsigned long)file->cluster_pointer;						// calculate the sector to be read. 
		temp1 += (unsigned long)file->sector_index;			
		mmc_read_sector((unsigned long)temp1,FileBuffer);						// Read the volume bootrecord from mmc.			

		if(file->filesize > 511) 
		{
			cnt = 512;															// 512 bytes of data are available in the filebuffer.
			file->filesize-=512;												// decrement the number of available 
		}
		else
		{
			cnt = file->filesize;												// cnt databytes are available in the filebuffer.
			file->filesize = 0;													// end of file reached.
		}
	
		file->sector_index++;													// continue reading in next sector
	}
	return(cnt);
}


//________________________________________________________________________________________________________________________________________
// Funtion: 	FileGetchar(File *file);
// 
// Description:	This function reads and returns one character from the specified file. Is the end of the actual sector reached the
//				next sector of the cluster be read. Is the last sector of the cluster read the next cluster will be searched in FAT.
//________________________________________________________________________________________________________________________________________
int FileGetchar(File *file)
{	
	int c = EOF;
	unsigned long temp1; 										
																				// is the sector in the buffer still up to date?
	if(file->filesize > 0)														// wen the end of the file is not reached, get the next character.
	{
		temp1  = (unsigned long)file->cluster_pointer;
		temp1 += (unsigned long)file->sector_index;			

		if(file->sector_in_buffer != temp1)	
		{
			mmc_read_sector((unsigned long)temp1,FileBuffer);					// Read the volume bootrecord from mmc.			
			file->sector_in_buffer = (unsigned long)temp1;
		}	
		c = FileBuffer[file->byte_index];
		file->filesize--;														// decrement the number of available 
	
		if(file->byte_index < 511)												// continue reading from this sector until the end of the sector is reached.
		{
			file->byte_index++;
		}
		else
		{
			file->byte_index=0;													// reading at the beginning of new sector.
			file->sector_index++;												// continue reading in next sector
			if(file->sector_index >= SectorsPerCluster)						// When end of cluster is reached, the next datacluster has to be searched in the FAT.
			{
				file->sector_index = 0;											// start reading new cluster at first sector of the cluster.
				GetNextCluster(file);											// Sets the clusterpointer of the file to the next datacluster.
			}
		}
	}
	return(c);
}

//-------------------------------------------------
char FileGetString(char *text, File *file, unsigned char count)
{
  
  int c='0';
  unsigned char i = 0;
  while ((c != EOF))
  { c = FileGetchar(file);
    
	//printf("%02X ",c);
    text[i++] = (unsigned char) c;
	if (c == '\n') break;
	if (i == count) break;
  }
  text[i-1] = '\0';
  if (c == EOF) return (0); else return(1);

}






//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int CreateFileInDirectory(unsigned char *, unsigned int)
// 
// Description:	This function looks for the next free position in the directory and creates a file with the specified filename and enters the
//				pointer to the first datacluster.
//________________________________________________________________________________________________________________________________________

unsigned char CreateFileInDirectory(unsigned char *fname, unsigned int cluster, File *fpoint)
{
	unsigned int 	rootentry	= 0;													// index to an entry in the rootdirectory.
	unsigned int 	cnt_enries_searched = 0;											// count the number of rootentries which have been searched already.
	unsigned char 	i = 0;
	unsigned int 	sector_offset = 0;													// index to the sector of the Rootentry which is searched momentarily
	unsigned char 	retvalue = 0;
																						// directory starts at sector specified by dir_sector. This can be the rootdirectory or any other directory.
//	printf("CreateFileInDirectory:\n\r");
	do
	{																					// search the next 16 rootentries in this sector of the roordirectory.
		rootentry=0;									
		mmc_read_sector((unsigned long)(RootDirectory + sector_offset),&FileBuffer[0]);	// Read the Rootdirectory.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		while((rootentry<16) && (!retvalue))
		{
			if(DirectoryEntry[rootentry].attribute == 0)								// empty directory entry found
			{
				for(i=0;i<11;i++) DirectoryEntry[rootentry].name[i] = fname[i];		// Kopie the filename and the file extension to the directoryentry.
				DirectoryEntry[rootentry].attribute    = 0x20;							// Set the fileattribute to archive to reserve the directoryentry.
				DirectoryEntry[rootentry].startcluster = cluster;						// copy the location of the first datacluster to the directoryentry.
				DirectoryEntry[rootentry].size     = 0;									// the new createted file has no content yet.
				fpoint->directory_sector = (unsigned long) (RootDirectory + sector_offset);
				fpoint->directory_index  = (unsigned char) rootentry;
				retvalue = 1;
//				printf("Cluster: %u \n\r",cluster);
				mmc_write_sector((unsigned long)(RootDirectory + sector_offset),&FileBuffer[0]);				
			}			
			rootentry++;
			cnt_enries_searched++;
		}
		if(!retvalue)																	// file not found in this sector so take next sector.
		{
			rootentry = 0;
			sector_offset++;
		}
	}
	while((cnt_enries_searched<PossibleRootEntries) && (!retvalue));

	return(retvalue);																	// return 1 if file has been created otherwise return 0.
}






//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int FindNextFreeCluster(void)
// 
// Description:	This function looks in the FAT to find the next free datacluster
//				
//________________________________________________________________________________________________________________________________________

unsigned int FindNextFreeCluster(File *file)
{
	unsigned long fat_pointer			=	0;							// Pointer to the first sector of the FAT.
	unsigned long ul_tmp				=  	0;							// temporary variable used to calculate a sectornumber.
	unsigned int  fat_sector_offset	=	0;							// index to a sector within the FAT.
	unsigned int  fat_entry			= 	0;							// index to an fatentry within the actual sector (256 fatentries are possible within one sector).
	unsigned int  free_cluster			= 	0;							// a pointer to the first sector of the next free cluster.
	
#if USE_MMC	
	fat_pointer = (unsigned long) FileAllocationTable;					// start searching for empty cluster at the beginning of the fat.
//	printf("FindNextFreeCluster:\n\r");																	// if the end of the fat is not reached yet and no free cluster has been found
	while((fat_sector_offset < SectorsPerFat) && (!free_cluster))	
	{
		ul_tmp = (unsigned long) ((unsigned long)fat_pointer + (unsigned long)fat_sector_offset);
		mmc_read_sector((unsigned long)ul_tmp,&FileBuffer[0]);				// read next sector of FAT.
		file->sector_in_buffer = ul_tmp;								// remember the number of the sector in FileBuffer.
		Fat = (struct FatEntry *)&FileBuffer[0];
		for(fat_entry=0;fat_entry<256;fat_entry++)						// look for an free cluster at all entries in this sector of the fat. 
		{
			if(Fat[fat_entry].next_cluster == 0x0000)					// empty cluster found!!
			{				
				Fat[fat_entry].next_cluster = 0xffff;					// mark this fat-entry as used and save it to the device.
//				printf("sect: %lu\n\r", file->sector_in_buffer);
				mmc_write_sector((unsigned long)file->sector_in_buffer,&FileBuffer[0]);
				free_cluster  = fat_entry;								// the relative position of the free cluster found in this sector of the FAT.
				free_cluster += (fat_sector_offset << 8);				// calculate the absolute position of the free cluster in the FAT;
				fat_entry = 256;										// terminate the search for a free cluster in this sector.
			}
		}
		fat_sector_offset++;									
	}
#endif	
	return(free_cluster);
}





//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int GetFatClusterIndex(File *file);
// 
// Description:	This function returns the clusterindex of the cluster specified by file->cluster_pointer of the specified file.
//				
//________________________________________________________________________________________________________________________________________

unsigned int GetFatClusterOffset(File *file)
{
	unsigned long fat_sector_offset	=	0;

	fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));				// Calculate index of actual cluster in FAT.
	fat_sector_offset  /= SectorsPerCluster;														
	fat_sector_offset  += 2;															// In Fat16 clusterpositions have an offset of two.								
	
	return((unsigned int)fat_sector_offset);
}




//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int GetFatSectorIndex(File *file);
// 
// Description:	This function returns the sectorindex of the cluster specified by file->cluster_pointer of the specified file.
//				
//________________________________________________________________________________________________________________________________________

unsigned int GetFatSectorIndex(File *file)
{
	unsigned int fat_pointer			=	0;
	
	fat_pointer = GetFatClusterOffset(file);
	fat_pointer = fat_pointer % 0x100;													// Calculate the clusterposition in the fat																
	
	return(fat_pointer);
}






//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int AppendCluster(File *file);
// 
// Description:	This function finds the next free datacluster on the disk and appends it to the specified file.
//				
//________________________________________________________________________________________________________________________________________
/*
unsigned int AppendCluster(File *file)
{
	unsigned int  free_cluster = 0;														
	unsigned long fat_pointer	= 0;
#if USE_MMC	
	
	free_cluster = FindNextFreeCluster(file);											// the next free cluster found on the disk.

	fat_pointer  = FileAllocationTable;													// Set Pointer to the beginnig of the FAT.
	fat_pointer += (unsigned long)((unsigned long)GetFatClusterOffset(file) >> 8);					// find the sector in the FAT with 256 entries per sector.
	
	Fat = (struct FatEntry *)&FileBuffer[0];
	Fat[GetFatSectorIndex(file)].next_cluster = free_cluster;							// append the free cluster to the end of the file in the FAT.
	mmc_write_sector((unsigned long)fat_pointer,&FileBuffer[0]);											// save the modified sector to the FAT.

	fat_pointer  = (unsigned long)free_cluster;
	fat_pointer -= 2;
	fat_pointer *= SectorsPerCluster;
	fat_pointer += FirstDataCluster;

	file->cluster_pointer = fat_pointer;												// continue wrtiting to the file in the new and free datacluster.
#endif
	return(1);
}
*/
static unsigned short AppendCluster(File *file) 
{
    unsigned short  free_cluster = 0;                                                        
    unsigned int fat_pointer    = 0;
#if USE_MMC	
  
  // the next free cluster found on the disk.
  // This free cluster will be marked as FFFF signalling that it is now the end of the file
    free_cluster = FindNextFreeCluster(file);                                            
  // Set Pointer to the beginnig of the FAT.
    fat_pointer  = FileAllocationTable;                                                    
  // find the correct sector in the FAT with 256 entries per sector.
    fat_pointer += GetFatClusterOffset(file) >> 8;
  // Read the sector we are now \"jumping\" out off..
    mmc_read_sector(fat_pointer,&FileBuffer[0]);                                            
    Fat = (struct FatEntry *)&FileBuffer[0];
  // append the free cluster to the end of the file in the FAT.
    Fat[GetFatSectorIndex(file)].next_cluster = free_cluster;                            
  // save the modified sector to the FAT.
    mmc_write_sector(fat_pointer,&FileBuffer[0]);                                            

    fat_pointer  = (unsigned int)free_cluster;
    fat_pointer -= 2;
    fat_pointer *= SectorsPerCluster;
    fat_pointer += FirstDataCluster;

  // continue writing to the file in the new and free datacluster.
  file->cluster_pointer = fat_pointer;
#endif

    return(1);

}


//________________________________________________________________________________________________________________________________________
// Funtion: 	void FindEndOfFile(void)
// 
// Description:	This function looks for the end of the specified file to appand data.
//				
//________________________________________________________________________________________________________________________________________
unsigned int FindEndOfFile(File *file)
{
	unsigned long	filesize;
	unsigned long temp; 										
#if USE_MMC	

	filesize = file->filesize;											// the size of the open file.

	while(filesize)
	{
		filesize--;														// decrement the number of available 	
		if(file->byte_index < 511)										// continue reading from this sector until the end of the sector is reached.
		{
			file->byte_index++;
		}
		else
		{
			file->byte_index=0;											// reading at the beginning of new sector.
			file->sector_index++;										// continue reading in next sector
			if(file->sector_index >= SectorsPerCluster)				// When end of cluster is reached, the next datacluster has to be searched in the FAT.
			{
				if(filesize) 
				{
			   	  file->sector_index = 0;									// start reading new cluster at first sector of the cluster.
				  GetNextCluster(file);						// Sets the clusterpointer of the file to the next datacluster.
				}
				/*else 		  
				{
					AppendCluster(file);								// append a new and free cluster at the end of the file.
				}*/	
			}
		}	
	}
	
	if(file->byte_index)				
	{ temp = (unsigned long)((unsigned long)file->cluster_pointer + (unsigned long)file->sector_index);			
	  mmc_read_sector((unsigned long)temp,&FileBuffer[0]);					// read FileBuffer 
	}  
#endif
	return(1);
}





//________________________________________________________________________________________________________________________________________
// Funtion: 	FilePutchar(File *file);
// 
// Description:	This function writes a byte to the specified file and takes care of writing the necessary FAT- Entries. 
//				
//________________________________________________________________________________________________________________________________________
unsigned char FilePutChar(File *file,char c)
{	
	unsigned long ul_temp = 0;
	
    if(file->sector_index >= SectorsPerCluster)									// if end of the cluster is reached, find next free cluster
	{
		file->sector_index = 0;
		AppendCluster(file);														// append a new and free cluster at the end of the file.
	}


	FileBuffer[file->byte_index] = c;														// write databyte into the buffer. The byte will be written to the device at once
	file->filesize++;																	// a character has been written to the file so the size is inkremented
																						// if the buffer contains the complete sectordata.
	if(file->byte_index < 511)															// if the end of this sector is not reached yet
	{
		file->byte_index++;																// the next byte will be written to the next byteposition in this sector.
	}
	else																				// otherwise the data in the sectorbuffer will be written to the device and the next sector will be selected.
	{
		ul_temp  = (unsigned long)file->cluster_pointer;
		ul_temp += (unsigned long)file->sector_index;
		
		mmc_write_sector((unsigned long)ul_temp,&FileBuffer[0]);
		file->byte_index=0;																// and the next byte will be written at the beginning of this new sector.
		file->sector_index++;
	}
	return(0);
}

unsigned char FileWriteSector(File *file,char *buf)
{	
	unsigned long ul_temp = 0;
	
		file->filesize+=512;
		
		ul_temp  = (unsigned long)file->cluster_pointer;
		ul_temp += (unsigned long)file->sector_index;
		
		mmc_write_sector((unsigned long)ul_temp,buf);
		file->byte_index=0;																// and the next byte will be written at the beginning of this new sector.
		file->sector_index++;
		if(file->sector_index >= SectorsPerCluster)									// if end of the cluster is reached, find next free cluster
		{
			file->sector_index = 0;
			AppendCluster(file);														// append a new and free cluster at the end of the file.
		}
	
	return(0);
}




//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int SeekFileInDirectory(unsigned char*, unsigned long, File *)
// 
// Description:	this function searches all possible rootentries until the file is found or the end of the rootdirectory is reached
//				
//________________________________________________________________________________________________________________________________________

unsigned int SeekFileInDirectory(unsigned char *fname,unsigned long dir_sector, File *fpoint)
{
	unsigned int cluster=0;								// Start search for file in the first sector of the rootdirectory.
	unsigned int rootentry=0;
	unsigned int cnt_enries_searched=0;
	unsigned char i=0;
	unsigned int  sector_offset=0;							// index to the sector of the Rootentry which is searched momentarily
															// directory starts at sector specified by dir_sector. This can be the rootdirectory or any other directory.
	#if USE_MMC	
	do
	{														// search the next 16 rootentries in this sector of the roordirectory.
		rootentry=0;									
		mmc_read_sector((unsigned long)((unsigned long)RootDirectory + (unsigned long)sector_offset),&FileBuffer[0]);		// Read the Rootdirectory.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		while((!cluster)&&(rootentry<16))
		{
			i=0;
			while((i<10)&&(DirectoryEntry[rootentry].name[i] == fname[i]))
			{
				i++;
			}
			if(i==10)											// file found!! -> reading startcluster of file from offset 26.
			{
				cluster = DirectoryEntry[rootentry].startcluster; 				
				fpoint->filesize = (unsigned long) DirectoryEntry[rootentry].size;
				fpoint->directory_sector = (unsigned long) (RootDirectory + sector_offset);
				fpoint->directory_index  = (unsigned char) rootentry;
			}
			rootentry++;
			cnt_enries_searched++;
		}
		if(!cluster)											// file not found in this sector so take next sector.
		{
			sector_offset++;
		}
	}
	while((cnt_enries_searched<PossibleRootEntries) && (!cluster));
#endif
	return(cluster);
}




//________________________________________________________________________________________________________________________________________
// Funtion: 	void SeperateFileName(unsigned char*);
// 
// Description:	This function seperates the filename and the fileattribute and brings them into the needed format ('test.txt' -> 'TEST    TXT');
//				
//________________________________________________________________________________________________________________________________________

void SeperateFileName(unsigned char *fname, unsigned char *name)
{
	unsigned char readpointer	= 0;
	unsigned char writepointer	= 0;

	while((writepointer<=10) && (fname[readpointer]!=0))	// the rootdirectoryentry is 8bytes for filename and 3bytes for fileattribute.
	{														// the filename in the rootdirectory is in the format "TEST    TXT" without the dot.
		if(fname[readpointer]=='.')						// seperating filename and attribute.
		{
			readpointer++;
			writepointer=8;			
		}
		else
		{
			if((fname[readpointer]>96) && (fname[readpointer]<123))
			{
				name[writepointer] = fname[readpointer] - 32;					// all characters must be upper case.
			}
			else name[writepointer]=fname[readpointer];
			
			readpointer++;
			writepointer++;
		}
	}
}




//________________________________________________________________________________________________________________________________________
// Funtion: 	FileClose(File *file);
// 
// Description:	This function closes the open file by writing the remaining data from the buffer to the device and entering the filesize 
//				in the directory entry.
//________________________________________________________________________________________________________________________________________

void FileClose(File *file)
{
#if USE_MMC	
/*
	if(file->mode =='a')
	{
		mmc_write_sector((unsigned long)(file->cluster_pointer + file->sector_index),&FileBuffer[0]);	// save the data still in the buffer 	
		mmc_read_sector((unsigned long)file->directory_sector,&FileBuffer[0]);								// read the directoryentry for this file.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		DirectoryEntry[file->directory_index].size = (unsigned long) file->filesize;
		mmc_write_sector((unsigned long)file->directory_sector,&FileBuffer[0]);
	}
*/
  if(file && file->mode =='a') {
    // If we have any data in the current FileBuffer (from FilePutChar()) we flush them to the disk first) 
    // This fixes the error that occurs if we call FilePutChar() with a number of bytes that will exactly fill the
    // current cluster we are writing to, and FilePutChar() calls AppendCluster(), this will leave the FileBuffer with FAT data in
    // hence we shold NOT write thoses data (FilePutChar() already wrote the correct data before it called AppendCluster())
    if (file->byte_index > 0) {
      // save the data still in the buffer 
      mmc_write_sector((unsigned int)(file->cluster_pointer + file->sector_index),&FileBuffer[0]); 
    }
    // read and update the directory entry for this file
    mmc_read_sector((unsigned int)file->directory_sector,&FileBuffer[0]); 
    DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
    DirectoryEntry[file->directory_index].size = (unsigned int) file->filesize;
    mmc_write_sector((unsigned int)file->directory_sector,&FileBuffer[0]);
  }
  file=0;
	
	
	
#endif
}



//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned char fopen(unsigned char*, unsigned char *, File *);
// 
// Description:	This function looks for the specified file in the rootdirectory of the drive. If the file is found the number of the 
//				corrosponding datacluster is returned to main. Only modes 'r' (reading) and 'a' append are implemented yet.
//
// Return:      0 = open failed 
//              1 = SUCCESS
//________________________________________________________________________________________________________________________________________

unsigned char FileOpen(unsigned char *fname, char mode, File *fpoint)
{
	unsigned long cluster_temp	= 0;
	unsigned char name[11] 	= "           ";			
	unsigned int temp;
	
	if (fpoint == NULL) return(0);                                                  // Valid pointer ??
	
	fpoint->start_cluster 		= 0;												// File not opend yet.
	fpoint->cluster_pointer 		= 0;
	fpoint->sector_index 		= 0;
	fpoint->byte_index			= 0;
	fpoint->mode					= mode;
	fpoint->filesize				= 0;

	SeperateFileName(&fname[0], &name[0]);											// seperate the filename and attribute and bring them in the correct format.			
	cluster_temp = (unsigned long)SeekFileInDirectory(&name[0],(unsigned long)RootDirectory, fpoint);
	if(cluster_temp)																// if file was found
	{
		temp = (unsigned int)cluster_temp;
		cluster_temp -=2;															// Clusterposition is ((position in FAT)-2). first two entries in FAT are reserved.
		cluster_temp *= (unsigned long)SectorsPerCluster;							// Calculate positon of first cluster.
		fpoint->start_cluster   = (FirstDataCluster + cluster_temp);
		fpoint->cluster_pointer = fpoint->start_cluster;								// start reading the file with the first cluster.
		if(mode == 'a')																// open existing file for writing (append data at the end of the file)
		{	
			FindEndOfFile(fpoint);
		}
		return(1);
	}
	else
	{
		if(mode == 'a')																// specified file doesn't exist so create new file for writing data.
		{	
			cluster_temp = (unsigned long)FindNextFreeCluster(fpoint);			// the next free cluster on the disk.
			if(cluster_temp) 														// if a free cluster is available:
			{
				temp = (unsigned int)cluster_temp;									// remember the index of the free datacluster found for the directory entry.
				cluster_temp -=2;													// Clusterposition is ((position in FAT)-2). first two entries in FAT are reserved.
				cluster_temp *= SectorsPerCluster;									// Calculate relative sectorindex of first datacluster.
				fpoint->start_cluster   = (FirstDataCluster + cluster_temp);			// Calculate absolute sectorposition of first datacluster.
				fpoint->cluster_pointer = fpoint->start_cluster;						// start reading the file with the first sector of the first datacluster.
				if(CreateFileInDirectory(&name[0],temp, fpoint))							// Could an entry for the new file in the rootdirectory be created?
				{
					return(1);
				}	
			}				
		}
		return(0);
	}	
}


#endif
